(*
 * Simple compile-time fault injector
 * Feng Zhou, 8/16/06
 *)

open Pretty
open Cil
open String
open Unix
module E = Errormsg
module H = Hashtbl

type workmode =
    ScanMode
    | PickMode
    | InjectMode
    | SkipMode;;

type faultkind =
    FlipCondition
    | LoopBound
    | ScanOverrun
    | OffByOne
    | CorruptParameter
    | MissingAssignment
    | MissingCall
    | UnknownKind
    ;;

let kind_of_string s =
  match s with
      "flipcondition" -> FlipCondition
    | "loopbound" -> LoopBound
    | "scanoverrun" -> ScanOverrun
    | "offbyone" -> OffByOne
    | "corruptparameter" -> CorruptParameter
    | "missingassignment" -> MissingAssignment
    | "missingcall" -> MissingCall
    | _ -> E.s (E.error "Unknown fault category %s." s)

let string_of_kind k =
  match k with
    FlipCondition -> "flipcondition"
    | LoopBound -> "loopbound"
    | ScanOverrun -> "scanoverrun"
    | OffByOne -> "offbyone"
    | CorruptParameter -> "corruptparameter"
    | MissingAssignment -> "missingassignment"
    | MissingCall -> "missingcall"
    | UnknownKind -> "unknown"

(* A list of memcpy-like functions that read or writes a buffer:
   function name -> index of size parameter *)
let scan_funcs : (string, int) H.t = H.create 10
let init_scan_funcs () =
  H.add scan_funcs "memcpy" 2;
  H.add scan_funcs "memset" 2;
  H.add scan_funcs "memchr" 2;
  H.add scan_funcs "memscan" 2;
  H.add scan_funcs "copy_from_user" 2;
  H.add scan_funcs "copy_to_user" 2
;;

let mode = ref SkipMode

type fileStat = {
    mutable candidates : int;
    mutable chosen : (int, int) H.t;
  }

(*
 * Handling the .ctswifi file
 *)
class dotCtswifi = object

  val mutable kind = FlipCondition
  val h : (string, fileStat) H.t = H.create 100

  (* kind of fault we care about *)
  method get_kind = kind
  method set_kind k = kind <- k

  (* add a new file if it does not exist *)
  method new_file file =
    let stat = {candidates=0; chosen=H.create 5} in
      H.replace h file stat

  (* get number of candidate for a file *)
  method get_candidates file =
    let stat = H.find h file in
      stat.candidates

  (* set new number of candidates *)
  method set_candidates file cand =
    let stat = H.find h file in
      stat.candidates <- cand;
      ()

  (* test if a candidate is chosen *)
  method is_chosen file c =
    let stat = H.find h file in
      H.mem stat.chosen c

  (* pick a number of locations to inject faults.  return actual
     number of locations flagged (could be smaller than requested if
     not enough candidates are available) *)
  method pick_faults nr =
    let ranges = Array.make ((H.length h)+1) (0,"") in
      (* clear chosen locations *)
      H.iter (fun file stat -> H.clear stat.chosen) h;
      (* calc 'ranges' storing index range for each file *)
    let (_, total) = H.fold
      (fun file stat (i, t) ->
	ranges.(i) <- (fst(ranges.(i)), file);
	ranges.(i+1) <- (t+stat.candidates, "");
	(i+1, t+stat.candidates) )
      h (0,0) in
      (* binary search to locate the file containing the fault indexed
	 at idx *)
    let rec find_file idx l r =
      let m = (l+r)/2 in
	if idx >= fst(ranges.(m)) && idx < fst(ranges.(m+1)) then
	  m
	else if idx >= fst(ranges.(m)) then
	  find_file idx (m+1) r
	else
	  find_file idx l (m-1)
    in
    let nr' = if nr < total then nr else total in
    let fin = ref 0 in
      while !fin < nr' do
	let p = Random.int total in
	let i = find_file p 0 (H.length h) in
	let (start,file) = ranges.(i) in
	let stat = H.find h file in
	  if not (H.mem stat.chosen (p-start)) then begin
	      H.add stat.chosen (p-start) (p-start);
	      fin := !fin+1
	    end
	  else
	    ()
      done;
      nr'

  (* load from .ctswifi *)
  method load dotFile =
    let nr = ref 0 in
      try
	let c = open_in dotFile in
	  kind <- kind_of_string (input_line c);
	  try
	    while true do
	      let line = input_line c in
	      let ss = Str.split (Str.regexp ":") line in
	      let file = List.hd ss in
	      let ss2 = if (List.length ss) > 2 then
		  Str.split (Str.regexp " ")  (List.nth ss 2)
		else [] in
	      let chosen = List.map int_of_string ss2  in
	      let stat = {
		  candidates = int_of_string (List.nth ss 1);
		  chosen = H.create 10
		} in
		ignore(List.map (fun x -> H.add stat.chosen x x) chosen);
		H.replace h file stat;
		nr := !nr + 1
	    done
	  with End_of_file ->
	    close_in c;
(*	    E.log "Loaded %d files\n" !nr *)
      with Sys_error(msg) ->
	  ()

  (* save to .ctswifi *)
  method save dotFile =
    let c = open_out dotFile in
      ignore(fprintf c "%s\n" (string_of_kind kind));
      H.iter
	(fun file stat ->
	  ignore(fprintf c "%s:%d:" file stat.candidates);
	  H.iter (fun x _ -> ignore(fprintf c "%d " x)) stat.chosen;
	  ignore(fprintf c "\n")
	) h;
      close_out c
end

let file = ref ""
let ctrl = new dotCtswifi
let flipCount = ref 0

(* increment fault counter for this file, or test if current fault
   should be inject when actually injecting*)
let flip = function () ->
  flipCount := !flipCount + 1;
  match !mode with
      ScanMode -> 
	ctrl#set_candidates !file !flipCount;
	false
    | InjectMode ->
	ctrl#is_chosen !file (!flipCount-1)
    | _ -> false

exception CannotChange;;

(* return a "corrupted" version of the expression of a specific type *)
let randomExp = function typ ->
  let typ = unrollType typ in
    match typ with
	TPtr(_,_) -> Formatcil.cExp "(void *) 0" []
      | TInt(ikind,_) -> begin
	    (* 0 with 0.5 prob, a random int with 0.5 prob *)
	    match Random.int 2 with
		0 -> Const (CInt64 (Int64.zero, ikind, None))
	      | _ -> Const (CInt64 (Int64.sub (Random.int64 4294967296L) 2147483648L, ikind, None))
	end
      | TFloat(fkind,_) -> begin
	    (* 0.0 with 0.5 prob, a random float between 0 and 1000
	       with 0.5 prob *)
	    match Random.int 2 with
		0 -> Const (CReal (0.0, fkind, None))
	      | _ -> Const (CReal ((Random.float 2000.0) -. 1000.0, fkind, None))
	end
      | TEnum(einfo, _) -> begin
	    (* Randomly pick an enum const *)
	    let (name, value, _) = List.nth einfo.eitems (Random.int (List.length einfo.eitems)) in
	      Const (CEnum (value, name, einfo))
	end
	  (* FIXME: what about TComp as function parameters? *)
      | _ -> E.log "Fault: cannot corrupt expression typed %a\n" d_type typ;
	  raise CannotChange
;;

class fVisitor = object
  inherit nopCilVisitor

  (* Only visit global funcitons coming from .c files ... *)
  method vglob (g: global) : global list visitAction =
    match g with
      | GFun(f,loc) ->
	  let s = sub loc.file ((length loc.file)-2) 2 in
	    if s = ".c" then
	      DoChildren
	    else begin
		SkipChildren
	      end
      | _ ->
	  SkipChildren

  method vblock (b: block) : block visitAction =
    if hasAttribute "trusted" b.battrs then
      SkipChildren			(* skip wrappers *)
    else
      DoChildren
end

(* flip all if conditions *)
class flipConditionVisitor = object
  inherit fVisitor

  method vstmt (s: stmt) : stmt visitAction =
    match s.skind with
	If(e,b1,b2,l) -> if flip () then (
	    E.log "Fault: flip if condition at %a\n" d_loc (get_stmtLoc s.skind);
	    let s' = mkStmt (If(UnOp(LNot,e,typeOf e), b1, b2, l)) in
	      ChangeDoChildrenPost (s', (fun x -> x))
	  ) else
	    DoChildren
      | _ -> DoChildren
end

let expandBound e =
  let delta = match Random.int 100 with
      x when x >= 0 && x < 50 -> 1
    | x when x >= 50 && x < 93 -> Random.int 1023 + 2
    | _ -> (Random.int 3 + 2) * 1024
  in
    BinOp (PlusA, e, Const (CInt64 (Int64.of_int delta,IInt,None)), typeOf e)

(* while(1) {...  if (i ...) break; where i is an integer. we make the
   count larger by: 50% 1 byte, 44% 2-1024 bytes, 6% 2-4KB *)
class loopBoundVisitor = object
  inherit fVisitor

  method vstmt (s: stmt) : stmt visitAction =
    match s.skind with
	(* Match anything like "if (var < ...) break;" *)
	If (BinOp (binop,(Lval(Var v, NoOffset) as lv),e2,typ), b1, b2, l) ->
	  if isIntegralType (typeOf e2) then begin
	      match b2.bstmts with
		  {skind=Break(_)} :: _ ->
		    if flip () then begin
			E.log "Fault: make loop bound at %a larger.\n" d_loc l;
			let s' = mkStmt (If(BinOp(binop,lv,expandBound e2,typ), b1, b2, l)) in
			  ChangeDoChildrenPost (s',(fun x -> x))
		      end
		    else
		      DoChildren
		| _ -> DoChildren
	    end
	  else
	    DoChildren
      | _ -> DoChildren
end

(* Off-by-one: change < to <= etc in expressions *)
class offByOneVisitor = object
  inherit fVisitor

  method vexpr (e: exp) : exp visitAction =
    match e with
	BinOp (op, e1, e2, t) ->
	  let b = match op with
	      Lt | Gt | Le | Ge -> true
	    | _ -> false
	  in
	    if b && flip () then
	      let op' = match op with
		  Lt -> Le
		| Gt -> Ge
		| Le -> Lt
		| _ -> Gt
	      in
		E.log "Fault: inserted off-by-one error in expression: %a\n" d_exp e;
		ChangeTo (BinOp (op', e1, e2, t))
	    else
	      DoChildren
      | _ -> DoChildren
end

(* Copy overrun: for calls to kmalloc(), memcpy(), copy_from_user(),
   copy_to_user(), we make the count larger by: 50% 1 byte, 44% 2-1024
   bytes, 6% 2-4KB. *)
class scanOverrunVisitor = object
  inherit fVisitor

  method vinst (i: instr) : instr list visitAction =
    match i with
	Call(_lv, (Lval (Var v, NoOffset) as func), args, l) ->
	  if H.mem scan_funcs v.vname && flip () then
	    let args_a = Array.of_list args in
	    let idx = H.find scan_funcs v.vname in
	      E.log "Fault: expanded size in call to %s() at %a\n" v.vname d_loc l;
	      args_a.(idx) <- expandBound args_a.(idx);
	      ChangeTo [Call(_lv, func, Array.to_list args_a, l)]
	  else
	    DoChildren
      | _ -> DoChildren
end

(* remove the result of assignment at random, this includes removing
   initializer of local/global variables. *)
class missingAssignmentVisitor = object
  inherit fVisitor

  method vinst (i: instr) : instr list visitAction =
    match i with
	Set(lv, e, l) ->
	  (* MissingAssignment (1/2) *)
	  if (flip ()) then (
	      E.log "Fault: deleted expression assignment of %a=%a at %a\n" d_lval lv
		d_exp e d_loc l;
	      ChangeTo []
	    ) else
	    DoChildren
      | Call(Some lv, func, args, l) ->
	  if flip () then
	    begin
	      (* MissingAssignment (2/2) *)
	      E.log "Ignored return value of function call at %a\n" d_loc l;
	      ChangeDoChildrenPost ([Call(None, func, args, l)], (fun x -> x))
	    end else DoChildren
      | _ -> DoChildren
end

(* remove a function call in an expression at random, if return value
   is non-pointer: the result is 50% 0, 50% a fixed random number. If
   pointer, NULL is used. *)
class missingCallVisitor = object
  inherit fVisitor

  method vinst (i: instr) : instr list visitAction =
    match i with
	Call(Some lv, func, args, l) ->
	  let typ = unrollType (typeOfLval lv) in begin
	    match typ with
		(* we cannot delete a call that returns a struct *)
		TComp(_,_) -> DoChildren
	      | _ ->
		  if flip () then begin
		      E.log "Fault: deleted call to %a at %a\n" d_exp func d_loc l;
		      ChangeTo [Set(lv, randomExp typ, l)]
		    end else
		    DoChildren
	    end
      | _ -> DoChildren
end

(* corrupt any non-pointer argument to a function at random, set any
   pointer parameter to NULL at random *)
class corruptParameterVisitor = object
  inherit fVisitor

  method vinst (i: instr) : instr list visitAction =
    match i with
	Call(Some lv, func, args, l) ->
	  if (List.length args > 0 && flip ()) then
	    begin
	      let idx = Random.int (List.length args) in
	      let args_a = Array.of_list args in
	      let typ = unrollType (typeOf (List.nth args idx)) in
		try
	          args_a.(idx) <- randomExp typ;
		  E.log "Fault: corrupted parameter %d to %a at %a\n" idx d_exp func d_loc l;
		  ChangeTo [Call(Some lv, func, Array.to_list args_a, l)]
		with CannotChange ->
		  (E.log "Fault: ignored change to parameter.\n";
		   DoChildren)
	    end else DoChildren
      | _ -> DoChildren
end

let dotfile = ".ctswifi";;

let getVisitor kind = match kind with
    FlipCondition -> new flipConditionVisitor
  | LoopBound -> new loopBoundVisitor
  | ScanOverrun -> new scanOverrunVisitor
  | OffByOne -> new offByOneVisitor
  | CorruptParameter -> new corruptParameterVisitor
  | MissingAssignment -> new missingAssignmentVisitor
  | MissingCall -> new missingCallVisitor
  | _ -> E.error "Not implemented yet";
      new fVisitor

let scanFaults (f: file) (fault: string) =
  init_scan_funcs ();
  file := f.fileName;
  mode := ScanMode;
  ctrl#load dotfile;
  let kind = kind_of_string fault in
    ctrl#set_kind kind;
    let fvisit = getVisitor kind in
      ctrl#new_file f.fileName;
      visitCilFileSameGlobals fvisit f;
      ctrl#save dotfile

let pickFaults (faults: int) (seed: int) =
  Random.init seed;
  init_scan_funcs();
  ctrl#load dotfile;
  let r = ctrl#pick_faults faults in 
    ctrl#save dotfile;
    r

let injectFaults (f: file) (seed: int) =
  Random.init seed;
  init_scan_funcs();
  mode := InjectMode;
  file := f.fileName;
  ctrl#load dotfile;
  try
    (* test if the file is specified in .ctswifi *)
    ignore(ctrl#is_chosen !file 0);
    let fvisit = getVisitor ctrl#get_kind in
      visitCilFileSameGlobals fvisit f;
      f
  with Not_found ->
    E.s (E.error "Faults for file %s are not specified in .ctswifi\n" !file)

let kindarg = ref ""
let seedarg = ref 1

let feature : featureDescr =
  { fd_name = "fault";
    fd_enabled = ref true;
    fd_description = "Compile-time software-implemented fault injection: "^
      "insert random faults into the code";
    fd_extraopt = [
	("--faultscan", Arg.String (fun s -> mode := ScanMode; kindarg := s),
	"\008=fault Do the scan pass (1st pass) of fault injection. Specify fault category:\n\t"^
	  "flipcondition, loopbound, scanoverrun, offbyone, corruptparameter, missingassignment, missingcall");
	("--faultpick",Arg.Int (fun s -> mode := PickMode; seedarg := s),
	"\008=seed Do the pick pass (2nd pass) of fault injection.");
	("--faultinject",Arg.Int (fun s -> mode := InjectMode; seedarg := s),
	"\008=seed Do the injection pass (last pass) of fault injection. Note: seed > 0")
      ];
    fd_doit =
      (function (f: file) ->
	match !mode with
	    ScanMode ->
	      scanFaults f !kindarg
	  | PickMode ->
	      ignore (pickFaults 5 !seedarg)
	  | InjectMode ->
	      ignore (injectFaults f !seedarg)
	  | _ -> ()
      );
    fd_post_check = true;
  }
